import sys
from abc import abstractmethod
from typing import TextIO

from zrb.context.any_shared_context import AnySharedContext


class AnyContext(AnySharedContext):
    """Abstract base class for managing task contexts, logging, and rendering.

    This class provides methods for managing context-specific data such as
    attempt counts, logging, and rendering templates with additional data.
    Subclasses must implement all abstract methods.
    """

    @abstractmethod
    def set_attempt(self, attempt: int):
        """Sets the current attempt count.

        Args:
            attempt (int): The current attempt count.
        """
        pass

    @abstractmethod
    def set_max_attempt(self, max_attempt: int):
        """Sets the maximum number of attempts.

        Args:
            max_attempt (int): The maximum number of attempts allowed.
        """
        pass

    @abstractmethod
    def update_task_env(self, task_env: dict[str, str]):
        pass

    @abstractmethod
    def print(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
        plain: bool = False,
    ):
        """Prints values to the specified output stream.

        Args:
            *values (object): The values to be printed.
            sep (str, optional): Separator to use between values. Defaults to a space.
            end (str, optional): String appended after the last value. Defaults to a newline.
            file (TextIO, optional): The output stream to print to. Defaults to sys.stderr.
            flush (bool, optional): Whether to flush the output stream. Defaults to True.
            plain (bool, optional): Whether to use plain text. Defaults to False.
        """
        pass

    @abstractmethod
    def log_debug(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        """Logs debug-level messages.

        Args:
            *values (object): The values to be logged.
            sep (str, optional): Separator to use between values. Defaults to a space.
            end (str, optional): String appended after the last value. Defaults to a newline.
            file (TextIO, optional): The output stream to log to. Defaults to sys.stderr.
            flush (bool, optional): Whether to flush the output stream. Defaults to True.
        """
        pass

    @abstractmethod
    def log_info(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        """Logs info-level messages.

        Args:
            *values (object): The values to be logged.
            sep (str, optional): Separator to use between values. Defaults to a space.
            end (str, optional): String appended after the last value. Defaults to a newline.
            file (TextIO, optional): The output stream to log to. Defaults to sys.stderr.
            flush (bool, optional): Whether to flush the output stream. Defaults to True.
        """
        pass

    @abstractmethod
    def log_warning(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        """Logs warning-level messages.

        Args:
            *values (object): The values to be logged.
            sep (str, optional): Separator to use between values. Defaults to a space.
            end (str, optional): String appended after the last value. Defaults to a newline.
            file (TextIO, optional): The output stream to log to. Defaults to sys.stderr.
            flush (bool, optional): Whether to flush the output stream. Defaults to True.
        """
        pass

    @abstractmethod
    def log_error(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        """Logs error-level messages.

        Args:
            *values (object): The values to be logged.
            sep (str, optional): Separator to use between values. Defaults to a space.
            end (str, optional): String appended after the last value. Defaults to a newline.
            file (TextIO, optional): The output stream to log to. Defaults to sys.stderr.
            flush (bool, optional): Whether to flush the output stream. Defaults to True.
        """
        pass

    @abstractmethod
    def log_critical(
        self,
        *values: object,
        sep: str | None = " ",
        end: str | None = "\n",
        file: TextIO | None = sys.stderr,
        flush: bool = True,
    ):
        """Logs critical-level messages.

        Args:
            *values (object): The values to be logged.
            sep (str, optional): Separator to use between values. Defaults to a space.
            end (str, optional): String appended after the last value. Defaults to a newline.
            file (TextIO, optional): The output stream to log to. Defaults to sys.stderr.
            flush (bool, optional): Whether to flush the output stream. Defaults to True.
        """
        pass

    @abstractmethod
    def render_bool(self, template: str | bool) -> bool:
        """Renders a template string into boolean.

        Args:
            template (str | bool): The template string to be rendered or the value.

        Returns:
            bool: The rendered template as a boolean.
        """
        pass

    @abstractmethod
    def render_int(self, template: str | int) -> int:
        """Renders a template string into integer.

        Args:
            template (str | int): The template string to be rendered or the value.

        Returns:
            int: The rendered template as a integer.
        """
        pass

    @abstractmethod
    def render_float(self, template: str | float) -> float:
        """Renders a template string into float.

        Args:
            template (str | float): The template string to be rendered or the value.

        Returns:
            float: The rendered template as a float.
        """
        pass



import json
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from zrb.config.llm_rate_limitter import LLMRateLimiter, llm_rate_limitter
from zrb.context.any_context import AnyContext
from zrb.context.any_shared_context import AnySharedContext
from zrb.task.llm.error import extract_api_error_details
from zrb.task.llm.print_node import print_node
from zrb.task.llm.tool_wrapper import wrap_func, wrap_tool
from zrb.task.llm.typing import ListOfDict

if TYPE_CHECKING:
    from pydantic_ai import Agent, Tool
    from pydantic_ai.agent import AgentRun
    from pydantic_ai.models import Model
    from pydantic_ai.settings import ModelSettings
    from pydantic_ai.toolsets import AbstractToolset

    ToolOrCallable = Tool | Callable
else:
    ToolOrCallable = Any


def create_agent_instance(
    ctx: AnyContext,
    model: "str | Model | None" = None,
    system_prompt: str = "",
    model_settings: "ModelSettings | None" = None,
    tools: list[ToolOrCallable] = [],
    toolsets: list["AbstractToolset[Agent]"] = [],
    retries: int = 3,
    is_yolo_mode: bool | None = None,
) -> "Agent":
    """Creates a new Agent instance with configured tools and servers."""
    from pydantic_ai import Agent, Tool
    from pydantic_ai.tools import GenerateToolJsonSchema

    if is_yolo_mode is None:
        is_yolo_mode = False
    # Normalize tools
    tool_list = []
    for tool_or_callable in tools:
        if isinstance(tool_or_callable, Tool):
            tool_list.append(tool_or_callable)
            # Update tool's function
            tool = tool_or_callable
            tool_list.append(
                Tool(
                    function=wrap_func(tool.function, ctx, is_yolo_mode),
                    takes_ctx=tool.takes_ctx,
                    max_retries=tool.max_retries,
                    name=tool.name,
                    description=tool.description,
                    prepare=tool.prepare,
                    docstring_format=tool.docstring_format,
                    require_parameter_descriptions=tool.require_parameter_descriptions,
                    schema_generator=GenerateToolJsonSchema,
                    strict=tool.strict,
                )
            )
        else:
            # Turn function into tool
            tool_list.append(wrap_tool(tool_or_callable, ctx, is_yolo_mode))
    # Return Agent
    return Agent(
        model=model,
        system_prompt=system_prompt,
        tools=tool_list,
        toolsets=toolsets,
        model_settings=model_settings,
        retries=retries,
    )


def get_agent(
    ctx: AnyContext,
    agent_attr: "Agent | Callable[[AnySharedContext], Agent] | None",
    model: "str | Model | None",
    system_prompt: str,
    model_settings: "ModelSettings | None",
    tools_attr: (
        list[ToolOrCallable] | Callable[[AnySharedContext], list[ToolOrCallable]]
    ),
    additional_tools: list[ToolOrCallable],
    toolsets_attr: "list[AbstractToolset[Agent]] | Callable[[AnySharedContext], list[AbstractToolset[Agent]]]",  # noqa
    additional_toolsets: "list[AbstractToolset[Agent]]",
    retries: int = 3,
    is_yolo_mode: bool | None = None,
) -> "Agent":
    """Retrieves the configured Agent instance or creates one if necessary."""
    from pydantic_ai import Agent

    # Render agent instance and return if agent_attr is already an agent
    if isinstance(agent_attr, Agent):
        return agent_attr
    if callable(agent_attr):
        agent_instance = agent_attr(ctx)
        if not isinstance(agent_instance, Agent):
            err_msg = (
                "Callable agent factory did not return an Agent instance, "
                f"got: {type(agent_instance)}"
            )
            raise TypeError(err_msg)
        return agent_instance
    # Get tools for agent
    tools = list(tools_attr(ctx) if callable(tools_attr) else tools_attr)
    tools.extend(additional_tools)
    # Get Toolsets for agent
    tool_sets = list(toolsets_attr(ctx) if callable(toolsets_attr) else toolsets_attr)
    tool_sets.extend(additional_toolsets)
    # If no agent provided, create one using the configuration
    return create_agent_instance(
        ctx=ctx,
        model=model,
        system_prompt=system_prompt,
        tools=tools,
        toolsets=tool_sets,
        model_settings=model_settings,
        retries=retries,
        is_yolo_mode=is_yolo_mode,
    )


async def run_agent_iteration(
    ctx: AnyContext,
    agent: "Agent",
    user_prompt: str,
    history_list: ListOfDict,
    rate_limitter: LLMRateLimiter | None = None,
    max_retry: int = 2,
) -> "AgentRun":
    """
    Runs a single iteration of the agent execution loop.

    Args:
        ctx: The task context.
        agent: The Pydantic AI agent instance.
        user_prompt: The user's input prompt.
        history_list: The current conversation history.

    Returns:
        The agent run result object.

    Raises:
        Exception: If any error occurs during agent execution.
    """
    if max_retry < 0:
        raise ValueError("Max retry cannot be less than 0")
    attempt = 0
    while attempt < max_retry:
        try:
            return await _run_single_agent_iteration(
                ctx=ctx,
                agent=agent,
                user_prompt=user_prompt,
                history_list=history_list,
                rate_limitter=rate_limitter,
            )
        except BaseException:
            attempt += 1
            if attempt == max_retry:
                raise
    raise Exception("Max retry exceeded")


async def _run_single_agent_iteration(
    ctx: AnyContext,
    agent: "Agent",
    user_prompt: str,
    history_list: ListOfDict,
    rate_limitter: LLMRateLimiter | None = None,
) -> "AgentRun":
    from openai import APIError
    from pydantic_ai.messages import ModelMessagesTypeAdapter

    agent_payload = estimate_request_payload(agent, user_prompt, history_list)
    if rate_limitter:
        await rate_limitter.throttle(agent_payload)
    else:
        await llm_rate_limitter.throttle(agent_payload)

    async with agent:
        async with agent.iter(
            user_prompt=user_prompt,
            message_history=ModelMessagesTypeAdapter.validate_python(history_list),
        ) as agent_run:
            async for node in agent_run:
                # Each node represents a step in the agent's execution
                # Reference: https://ai.pydantic.dev/agents/#streaming
                try:
                    await print_node(_get_plain_printer(ctx), agent_run, node)
                except APIError as e:
                    # Extract detailed error information from the response
                    error_details = extract_api_error_details(e)
                    ctx.log_error(f"API Error: {error_details}")
                    raise
                except Exception as e:
                    ctx.log_error(f"Error processing node: {str(e)}")
                    ctx.log_error(f"Error type: {type(e).__name__}")
                    raise
            return agent_run


def estimate_request_payload(
    agent: "Agent", user_prompt: str, history_list: ListOfDict
) -> str:
    system_prompts = agent._system_prompts if hasattr(agent, "_system_prompts") else ()
    return json.dumps(
        [
            {"role": "system", "content": "\n".join(system_prompts)},
            *history_list,
            {"role": "user", "content": user_prompt},
        ]
    )


def _get_plain_printer(ctx: AnyContext):
    def printer(*args, **kwargs):
        if "plain" not in kwargs:
            kwargs["plain"] = True
        return ctx.print(*args, **kwargs)

    return printer



import json
import os
from collections.abc import Callable
from typing import Any

from zrb.config.llm_context.config import llm_context_config
from zrb.context.any_context import AnyContext
from zrb.task.llm.typing import ListOfDict
from zrb.util.file import read_file
from zrb.util.llm.prompt import make_prompt_section
from zrb.util.run import run_async


class ConversationHistory:

    def __init__(
        self,
        past_conversation_summary: str = "",
        past_conversation_transcript: str = "",
        history: ListOfDict | None = None,
        contextual_note: str | None = None,
        long_term_note: str | None = None,
        project_path: str | None = None,
    ):
        self.past_conversation_transcript = past_conversation_transcript
        self.past_conversation_summary = past_conversation_summary
        self.history = history if history is not None else []
        self.contextual_note = contextual_note if contextual_note is not None else ""
        self.long_term_note = long_term_note if long_term_note is not None else ""
        self.project_path = project_path if project_path is not None else os.getcwd()

    def to_dict(self) -> dict[str, Any]:
        return {
            "past_conversation_summary": self.past_conversation_summary,
            "past_conversation_transcript": self.past_conversation_transcript,
            "history": self.history,
            "contextual_note": self.contextual_note,
            "long_term_note": self.long_term_note,
        }

    def model_dump_json(self, indent: int = 2) -> str:
        return json.dumps(self.to_dict(), indent=indent)

    @classmethod
    async def read_from_source(
        cls,
        ctx: AnyContext,
        reader: Callable[[AnyContext], dict[str, Any] | list | None] | None,
        file_path: str | None,
    ) -> "ConversationHistory | None":
        # Priority 1: Reader function
        if reader:
            try:
                raw_data = await run_async(reader(ctx))
                if raw_data:
                    instance = cls.parse_and_validate(ctx, raw_data, "reader")
                    if instance:
                        return instance
            except Exception as e:
                ctx.log_warning(
                    f"Error executing conversation history reader: {e}. Ignoring."
                )
        # Priority 2: History file
        if file_path and os.path.isfile(file_path):
            try:
                content = read_file(file_path)
                raw_data = json.loads(content)
                instance = cls.parse_and_validate(ctx, raw_data, f"file '{file_path}'")
                if instance:
                    return instance
            except json.JSONDecodeError:
                ctx.log_warning(
                    f"Could not decode JSON from history file '{file_path}'. "
                    "Ignoring file content."
                )
            except Exception as e:
                ctx.log_warning(
                    f"Error reading history file '{file_path}': {e}. "
                    "Ignoring file content."
                )
        # Fallback: Return default value
        return None

    def fetch_newest_notes(self):
        self._fetch_long_term_note()
        self._fetch_contextual_note()

    @classmethod
    def parse_and_validate(
        cls, ctx: AnyContext, data: Any, source: str
    ) -> "ConversationHistory":
        try:
            if isinstance(data, cls):
                return data  # Already a valid instance
            if isinstance(data, dict):
                # This handles both the new format and the old {'context': ..., 'history': ...}
                return cls(
                    past_conversation_summary=data.get("past_conversation_summary", ""),
                    past_conversation_transcript=data.get(
                        "past_conversation_transcript", ""
                    ),
                    history=data.get("history", data.get("messages", [])),
                    contextual_note=data.get("contextual_note", ""),
                    long_term_note=data.get("long_term_note", ""),
                )
            elif isinstance(data, list):
                # Handle very old format (just a list) - wrap it
                ctx.log_warning(
                    f"History from {source} contains legacy list format. "
                    "Wrapping it into the new structure. "
                    "Consider updating the source format."
                )
                return cls(history=data)
            else:
                ctx.log_warning(
                    f"History data from {source} has unexpected format "
                    f"(type: {type(data)}). Ignoring."
                )
        except Exception as e:  # Catch validation errors too
            ctx.log_warning(
                f"Error validating/parsing history data from {source}: {e}. Ignoring."
            )
        return cls()

    def write_past_conversation_summary(self, past_conversation_summary: str):
        """
        Write or update the past conversation summary.

        Use this tool to store or update a summary of previous conversations for
        future reference. This is useful for providing context to LLMs or other tools
        that need a concise history.

        Args:
            past_conversation_summary (str): The summary text to store.

        Returns:
            str: A JSON object indicating the success or failure of the operation.

        Raises:
            Exception: If the summary cannot be written.
        """
        self.past_conversation_summary = past_conversation_summary
        return json.dumps({"success": True})

    def write_past_conversation_transcript(self, past_conversation_transcript: str):
        """
        Write or update the past conversation transcript.

        Use this tool to store or update the full transcript of previous conversations.
        This is useful for providing detailed context to LLMs or for record-keeping.

        Args:
            past_conversation_transcript (str): The transcript text to store.

        Returns:
            str: A JSON object indicating the success or failure of the operation.

        Raises:
            Exception: If the transcript cannot be written.
        """
        self.past_conversation_transcript = past_conversation_transcript
        return json.dumps({"success": True})

    def read_long_term_note(self) -> str:
        """
        Read the content of the long-term references.

        This tool helps you retrieve knowledge or notes stored for long-term reference.
        If the note does not exist, you may want to create it using the write tool.

        Returns:
            str: JSON with content of the notes.

        Raises:
            Exception: If the note cannot be read.
        """
        return json.dumps({"content": self._fetch_long_term_note()})

    def write_long_term_note(self, content: str) -> str:
        """
        Write the entire content of the long-term references.
        This will overwrite any existing long-term notes.

        Args:
            content (str): The full content of the long-term notes.

        Returns:
            str: JSON indicating success.
        """
        llm_context_config.write_context(content, context_path="/")
        return json.dumps({"success": True})

    def read_contextual_note(self) -> str:
        """
        Read the content of the contextual references for the current project.

        This tool helps you retrieve knowledge or notes stored for contextual reference.
        If the note does not exist, you may want to create it using the write tool.

        Returns:
            str: JSON with content of the notes.

        Raises:
            Exception: If the note cannot be read.
        """
        return json.dumps({"content": self._fetch_contextual_note()})

    def write_contextual_note(
        self, content: str, context_path: str | None = None
    ) -> str:
        """
        Write the entire content of the contextual references for a specific path.
        This will overwrite any existing contextual notes for that path.

        Args:
            content (str): The full content of the contextual notes.
            context_path (str, optional): The directory path for the context.
                                        Defaults to the current project path.

        Returns:
            str: JSON indicating success.
        """
        if context_path is None:
            context_path = self.project_path
        llm_context_config.write_context(content, context_path=context_path)
        return json.dumps({"success": True})

    def _fetch_long_term_note(self):
        contexts = llm_context_config.get_contexts(cwd=self.project_path)
        self.long_term_note = contexts.get("/", "")
        return self.long_term_note

    def _fetch_contextual_note(self):
        contexts = llm_context_config.get_contexts(cwd=self.project_path)
        self.contextual_note = "\n".join(
            [
                make_prompt_section(header, content)
                for header, content in contexts.items()
                if header != "/"
            ]
        )
        return self.contextual_note



import os

from zrb.input.str_input import StrInput

project_dir_input = StrInput(
    name="project-dir",
    description="Project directory",
    prompt="Project directory",
    default=lambda _: os.getcwd(),
)

app_name_input = StrInput(
    name="app",
    description="App name",
    prompt="App name",
)



from zrb.builtin.git import git_commit
from zrb.builtin.group import git_subtree_group
from zrb.context.any_context import AnyContext
from zrb.input.str_input import StrInput
from zrb.task.make_task import make_task
from zrb.util.cli.style import stylize_faint
from zrb.util.git import get_repo_dir
from zrb.util.git_subtree import add_subtree, load_config, pull_subtree, push_subtree


@make_task(
    name="git-add-subtree",
    input=[
        StrInput(
            name="name", description="Subtree config name", prompt="Subtree config name"
        ),
        StrInput(
            name="repo-url", description="Subtree repo URL", prompt="Subtree repo URL"
        ),
        StrInput(
            name="repo-branch",
            description="Subtree repo branch",
            prompt="Subtree repo branch",
        ),
        StrInput(
            name="repo-prefix",
            description="Subtree repo prefix",
            prompt="Subtree repo prefix",
        ),
    ],
    description="➕ Add subtree",
    upstream=git_commit,
    group=git_subtree_group,
    alias="add",
)
async def git_add_subtree(ctx: AnyContext):
    ctx.print(stylize_faint("Get directory"))
    repo_dir = await get_repo_dir(print_method=ctx.print)
    ctx.print(stylize_faint("Add subtree"))
    await add_subtree(
        repo_dir=repo_dir,
        name=ctx.input.name,
        repo_url=ctx.input["repo-url"],
        branch=ctx.input["repo-branch"],
        prefix=ctx.input["repo-prefix"],
        print_method=ctx.print,
    )


@make_task(
    name="git-pull-subtree",
    description="📥 Pull subtree",
    upstream=git_commit,
    group=git_subtree_group,
    alias="pull",
)
async def git_pull_subtree(ctx: AnyContext):
    ctx.print(stylize_faint("Get directory"))
    repo_dir = await get_repo_dir(print_method=ctx.print)
    config = load_config(repo_dir)
    if not config.data:
        raise ValueError("No subtree config found")
    first_err: Exception | None = None
    for name, detail in config.data.items():
        try:
            ctx.print(stylize_faint(f"Pull from subtree {name}"))
            await pull_subtree(
                repo_dir=repo_dir,
                prefix=detail.prefix,
                repo_url=detail.repo_url,
                branch=detail.branch,
                print_method=ctx.print,
            )
        except Exception as e:
            if first_err is None:
                first_err = e
            ctx.log_error(e)
    if first_err is not None:
        raise first_err


@make_task(
    name="git-push-subtree",
    description="📤 Push subtree",
    upstream=git_commit,
    group=git_subtree_group,
    alias="push",
)
async def git_push_subtree(ctx: AnyContext):
    ctx.print(stylize_faint("Get directory"))
    repo_dir = await get_repo_dir(print_method=ctx.print)
    config = load_config(repo_dir)
    if not config.data:
        raise ValueError("No subtree config found")
    first_err: Exception | None = None
    for name, detail in config.data.items():
        try:
            ctx.print(stylize_faint(f"Push to subtree {name}"))
            await push_subtree(
                repo_dir=repo_dir,
                prefix=detail.prefix,
                repo_url=detail.repo_url,
                branch=detail.branch,
                print_method=ctx.print,
            )
        except Exception as e:
            if first_err is None:
                first_err = e
            ctx.log_error(e)
    if first_err is not None:
        raise first_err



import functools
import inspect
import traceback
import typing
from collections.abc import Callable
from typing import TYPE_CHECKING

from zrb.context.any_context import AnyContext
from zrb.task.llm.error import ToolExecutionError
from zrb.util.callable import get_callable_name
from zrb.util.cli.style import (
    stylize_blue,
    stylize_error,
    stylize_green,
    stylize_yellow,
)
from zrb.util.run import run_async
from zrb.util.string.conversion import to_boolean

if TYPE_CHECKING:
    from pydantic_ai import Tool


def wrap_tool(func: Callable, ctx: AnyContext, is_yolo_mode: bool) -> "Tool":
    """Wraps a tool function to handle exceptions and context propagation."""
    from pydantic_ai import RunContext, Tool

    original_sig = inspect.signature(func)
    needs_run_context_for_pydantic = _has_context_parameter(original_sig, RunContext)
    wrapper = wrap_func(func, ctx, is_yolo_mode)
    return Tool(wrapper, takes_ctx=needs_run_context_for_pydantic)


def wrap_func(func: Callable, ctx: AnyContext, is_yolo_mode: bool) -> Callable:
    original_sig = inspect.signature(func)
    needs_any_context_for_injection = _has_context_parameter(original_sig, AnyContext)
    takes_no_args = len(original_sig.parameters) == 0
    # Pass individual flags to the wrapper creator
    wrapper = _create_wrapper(
        func, original_sig, ctx, needs_any_context_for_injection, is_yolo_mode
    )
    _adjust_signature(wrapper, original_sig, takes_no_args)
    return wrapper


def _has_context_parameter(original_sig: inspect.Signature, context_type: type) -> bool:
    """
    Checks if the function signature includes a parameter with the specified
    context type annotation.
    """
    return any(
        _is_annotated_with_context(param.annotation, context_type)
        for param in original_sig.parameters.values()
    )


def _is_annotated_with_context(param_annotation, context_type):
    """
    Checks if the parameter annotation is the specified context type
    or a generic type containing it (e.g., Optional[ContextType]).
    """
    if param_annotation is inspect.Parameter.empty:
        return False
    if param_annotation is context_type:
        return True
    # Check for generic types like Optional[ContextType] or Union[ContextType, ...]
    origin = typing.get_origin(param_annotation)
    args = typing.get_args(param_annotation)
    if origin is not None and args:
        # Check if context_type is one of the arguments of the generic type
        return any(arg is context_type for arg in args)
    return False


def _create_wrapper(
    func: Callable,
    original_sig: inspect.Signature,
    ctx: AnyContext,
    needs_any_context_for_injection: bool,
    is_yolo_mode: bool,
) -> Callable:
    """Creates the core wrapper function."""

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        # Identify AnyContext parameter name from the original signature if needed
        any_context_param_name = None
        if needs_any_context_for_injection:
            for param in original_sig.parameters.values():
                if _is_annotated_with_context(param.annotation, AnyContext):
                    any_context_param_name = param.name
                    break  # Found it, no need to continue
            if any_context_param_name is None:
                # This should not happen if needs_any_context_for_injection is True,
                # but check for safety
                raise ValueError(
                    "AnyContext parameter name not found in function signature."
                )
            # Inject the captured ctx into kwargs. This will overwrite if the LLM
            # somehow provided it.
            kwargs[any_context_param_name] = ctx
        # If the dummy argument was added for schema generation and is present in kwargs,
        # remove it before calling the original function, unless the original function
        # actually expects a parameter named '_dummy'.
        if "_dummy" in kwargs and "_dummy" not in original_sig.parameters:
            del kwargs["_dummy"]
        try:
            if not is_yolo_mode and not ctx.is_web_mode and ctx.is_tty:
                approval, reason = await _ask_for_approval(ctx, func, *args, **kwargs)
                if not approval:
                    raise ValueError(f"User disapproving: {reason}")
            return await run_async(func(*args, **kwargs))
        except Exception as e:
            error_model = ToolExecutionError(
                tool_name=func.__name__,
                error_type=type(e).__name__,
                message=str(e),
                details=traceback.format_exc(),
            )
            return error_model.model_dump_json()

    return wrapper


async def _ask_for_approval(
    ctx: AnyContext, func: Callable, *args, **kwargs
) -> tuple[bool, str]:
    func_name = get_callable_name(func)
    normalized_args = [stylize_green(_truncate_arg(arg)) for arg in args]
    normalized_kwargs = [
        f"{stylize_yellow(key)}={stylize_green(_truncate_arg(val))}"
        for key, val in kwargs.items()
    ]
    func_param_str = ",".join(normalized_args + normalized_kwargs)
    func_call_str = (
        f"{stylize_blue(func_name + '(')}{func_param_str}{stylize_blue(')')}"
    )
    while True:
        ctx.print(
            f"✅ >> Allow to run tool: {func_call_str} (Yes | No, <reason>)", plain=True
        )
        user_input = await _read_line()
        user_responses = [val.strip() for val in user_input.split(",", maxsplit=2)]
        while len(user_responses) < 2:
            user_responses.append("")
        approval_str, reason = user_responses
        try:
            approved = to_boolean(approval_str)
            if not approved and reason == "":
                ctx.print(
                    stylize_error(
                        f"You must specify rejection reason (i.e., No, <why>) for {func_call_str}"
                    ),  # noqa
                    plain=True,
                )
                continue
            return approved, reason
        except Exception:
            ctx.print(
                stylize_error(
                    f"Invalid approval value for {func_call_str}: {approval_str}"
                ),
                plain=True,
            )
            continue


def _truncate_arg(arg: str, length: int = 19) -> str:
    if len(arg) > length:
        return f"{arg[:length-4]} ..."
    return arg


async def _read_line():
    from prompt_toolkit import PromptSession

    reader = PromptSession()
    return await reader.prompt_async()


def _adjust_signature(
    wrapper: Callable, original_sig: inspect.Signature, takes_no_args: bool
):
    """Adjusts the wrapper function's signature for schema generation."""
    # The wrapper's signature should represent the arguments the *LLM* needs to provide.
    # The LLM does not provide RunContext (pydantic-ai injects it) or AnyContext
    # (we inject it). So, the wrapper's signature should be the original signature,
    # minus any parameters annotated with RunContext or AnyContext.

    from pydantic_ai import RunContext

    params_for_schema = [
        param
        for param in original_sig.parameters.values()
        if not _is_annotated_with_context(param.annotation, RunContext)
        and not _is_annotated_with_context(param.annotation, AnyContext)
    ]

    # If after removing context parameters, there are no parameters left,
    # and the original function took no args, keep the dummy.
    # If after removing context parameters, there are no parameters left,
    # but the original function *did* take args (only context), then the schema
    # should have no parameters.
    if not params_for_schema and takes_no_args:
        # Keep the dummy if the original function truly had no parameters
        new_sig = inspect.Signature(
            parameters=[
                inspect.Parameter(
                    "_dummy", inspect.Parameter.POSITIONAL_OR_KEYWORD, default=None
                )
            ]
        )
    else:
        new_sig = inspect.Signature(parameters=params_for_schema)

    wrapper.__signature__ = new_sig



from typing import Any

from zrb.context.shared_context import SharedContext
from zrb.task.any_task import AnyTask


def get_run_kwargs(
    task: AnyTask, args: list[str], kwargs: dict[str, str], cli_mode: bool
) -> dict[str, str]:
    arg_index = 0
    str_kwargs = {key: f"{val}" for key, val in kwargs.items()}
    run_kwargs = {**str_kwargs}
    shared_ctx = SharedContext(args=args)
    for task_input in task.inputs:
        if task_input.name in str_kwargs:
            # Update shared context for next input default value
            task_input.update_shared_context(shared_ctx, str_kwargs[task_input.name])
        elif arg_index < len(args) and task_input.allow_positional_parsing:
            run_kwargs[task_input.name] = args[arg_index]
            # Update shared context for next input default value
            task_input.update_shared_context(shared_ctx, run_kwargs[task_input.name])
            arg_index += 1
        else:
            if cli_mode and task_input.always_prompt:
                str_value = task_input.prompt_cli_str(shared_ctx)
            else:
                str_value = task_input.get_default_str(shared_ctx)
            run_kwargs[task_input.name] = str_value
            # Update shared context for next input default value
            task_input.update_shared_context(shared_ctx, run_kwargs[task_input.name])
    return run_kwargs



from my_app_name.module.gateway.schema.navigation import Navigation, Page, PageGroup

APP_NAVIGATION = Navigation()

APP_NAVIGATION.append_page(Page(name="gateway.home", caption="Home", url="/"))

auth_menu = APP_NAVIGATION.append_page_group(
    PageGroup(
        name="auth",
        caption="Authorization",
    )
)

auth_menu.append_page(
    Page(
        name="auth.permission",
        caption="Permission",
        url="/auth/permissions",
        permission="permission:read",
    )
)

auth_menu.append_page(
    Page(
        name="auth.role",
        caption="Role",
        url="/auth/roles",
        permission="role:read",
    )
)

auth_menu.append_page(
    Page(
        name="auth.user",
        caption="User",
        url="/auth/users",
        permission="user:read",
    )
)



from typing import Annotated

from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.exception_handlers import http_exception_handler
from fastapi.responses import HTMLResponse, RedirectResponse
from my_app_name.common.app_factory import app
from my_app_name.common.schema import BasicResponse
from my_app_name.config import (
    APP_AUTH_ACCESS_TOKEN_COOKIE_NAME,
    APP_MAIN_MODULE,
    APP_MODE,
    APP_MODULES,
)
from my_app_name.module.auth.client.auth_client_factory import auth_client
from my_app_name.module.gateway.subroute.auth import serve_auth_route
from my_app_name.module.gateway.util.auth import get_current_user
from my_app_name.module.gateway.util.view import render_content, render_error
from my_app_name.schema.user import AuthUserResponse


def serve_route(app: FastAPI):
    if APP_MODE != "monolith" and "gateway" not in APP_MODULES:
        return
    if APP_MODE == "monolith" or APP_MAIN_MODULE == "gateway":
        _serve_health_check(app)
        _serve_readiness_check(app)
        _serve_common_pages(app)
        _handle_404(app)

    # Serve auth routes
    serve_auth_route(app)


def _serve_common_pages(app: FastAPI):
    @app.get("/", include_in_schema=False)
    def home_page(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    ):
        return render_content(
            view_path="homepage.html",
            current_user=current_user,
            page_name="gateway.home",
        )

    @app.get("/login", include_in_schema=False)
    def login_page(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    ):
        if not current_user.is_guest:
            return RedirectResponse("/")
        return render_content(
            view_path="login.html",
            current_user=current_user,
            page_name="gateway.home",
            partials={"show_user_info": False},
        )

    @app.get("/logout", include_in_schema=False)
    def logout_page(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    ):
        if current_user is None:
            return RedirectResponse("/")
        return render_content(
            view_path="logout.html",
            current_user=current_user,
            page_name="gateway.home",
            partials={"show_user_info": False},
        )


def _serve_health_check(app: FastAPI):
    @app.api_route("/health", methods=["GET", "HEAD"], response_model=BasicResponse)
    async def health():
        """
        My App Name's health check
        """
        return BasicResponse(message="ok")


def _serve_readiness_check(app: FastAPI):
    @app.api_route("/readiness", methods=["GET", "HEAD"], response_model=BasicResponse)
    async def readiness():
        """
        My App Name's readiness check
        """
        return BasicResponse(message="ok")


def _handle_404(app: FastAPI):
    @app.exception_handler(404)
    async def default_404(request: Request, exc: HTTPException) -> HTMLResponse:
        if request.url.path.startswith("/api"):
            # Re-raise the exception to let FastAPI handle it
            return await http_exception_handler(request, exc)
        # Get current user by cookies
        current_user = None
        cookie_access_token = request.cookies.get(APP_AUTH_ACCESS_TOKEN_COOKIE_NAME)
        if cookie_access_token is not None and cookie_access_token != "":
            current_user = await auth_client.get_current_user(cookie_access_token)
        # Show error page
        return render_error(
            error_message="Not found", status_code=404, current_user=current_user
        )


serve_route(app)



import os
import platform
import re
from datetime import datetime, timezone

from zrb.attr.type import StrAttr, StrListAttr
from zrb.config.llm_config import llm_config as llm_config
from zrb.config.llm_context.config import llm_context_config
from zrb.context.any_context import AnyContext
from zrb.task.llm.conversation_history_model import ConversationHistory
from zrb.util.attr import get_attr, get_str_attr, get_str_list_attr
from zrb.util.file import read_dir, read_file_with_line_numbers
from zrb.util.llm.prompt import make_prompt_section


def get_persona(
    ctx: AnyContext,
    persona_attr: StrAttr | None,
    render_persona: bool,
) -> str:
    """Gets the persona, prioritizing task-specific, then default."""
    persona = get_attr(
        ctx,
        persona_attr,
        None,
        auto_render=render_persona,
    )
    if persona is not None:
        return persona
    return llm_config.default_persona or ""


def get_base_system_prompt(
    ctx: AnyContext,
    system_prompt_attr: StrAttr | None,
    render_system_prompt: bool,
) -> str:
    """Gets the base system prompt, prioritizing task-specific, then default."""
    system_prompt = get_attr(
        ctx,
        system_prompt_attr,
        None,
        auto_render=render_system_prompt,
    )
    if system_prompt is not None:
        return system_prompt
    return llm_config.default_system_prompt or ""


def get_special_instruction_prompt(
    ctx: AnyContext,
    special_instruction_prompt_attr: StrAttr | None,
    render_spcecial_instruction_prompt: bool,
) -> str:
    """Gets the special instruction prompt, prioritizing task-specific, then default."""
    special_instruction = get_attr(
        ctx,
        special_instruction_prompt_attr,
        None,
        auto_render=render_spcecial_instruction_prompt,
    )
    if special_instruction is not None:
        return special_instruction
    return llm_config.default_special_instruction_prompt


def get_modes(
    ctx: AnyContext,
    modes_attr: StrListAttr | None,
    render_modes: bool,
) -> str:
    """Gets the modes, prioritizing task-specific, then default."""
    raw_modes = get_str_list_attr(
        ctx,
        [] if modes_attr is None else modes_attr,
        auto_render=render_modes,
    )
    if raw_modes is None:
        raw_modes = []
    modes = [mode.strip() for mode in raw_modes if mode.strip() != ""]
    if len(modes) > 0:
        return modes
    return llm_config.default_modes or []


def get_workflow_prompt(
    ctx: AnyContext,
    modes_attr: StrListAttr | None,
    render_modes: bool,
) -> str:
    modes = get_modes(ctx, modes_attr, render_modes)
    # Get user-defined workflows
    workflows = {
        workflow_name: content
        for workflow_name, content in llm_context_config.get_workflows().items()
        if workflow_name in modes
    }
    # Get requested builtin-workflow names
    requested_builtin_workflow_names = [
        workflow_name
        for workflow_name in ("coding", "copywriting", "researching")
        if workflow_name in modes and workflow_name not in workflows
    ]
    # add builtin-workflows if requested
    if len(requested_builtin_workflow_names) > 0:
        dir_path = os.path.dirname(__file__)
        for workflow_name in requested_builtin_workflow_names:
            workflow_file_path = os.path.join(
                dir_path, "default_workflow", f"{workflow_name}.md"
            )
            with open(workflow_file_path, "r") as f:
                workflows[workflow_name] = f.read()
    return "\n".join(
        [
            make_prompt_section(header.capitalize(), content)
            for header, content in workflows.items()
            if header.lower() in modes
        ]
    )


def get_system_and_user_prompt(
    ctx: AnyContext,
    user_message: str,
    persona_attr: StrAttr | None = None,
    render_persona: bool = False,
    system_prompt_attr: StrAttr | None = None,
    render_system_prompt: bool = False,
    special_instruction_prompt_attr: StrAttr | None = None,
    render_special_instruction_prompt: bool = False,
    modes_attr: StrListAttr | None = None,
    render_modes: bool = False,
    conversation_history: ConversationHistory | None = None,
) -> tuple[str, str]:
    """Combines persona, base system prompt, and special instructions."""
    persona = get_persona(ctx, persona_attr, render_persona)
    base_system_prompt = get_base_system_prompt(
        ctx, system_prompt_attr, render_system_prompt
    )
    special_instruction_prompt = get_special_instruction_prompt(
        ctx, special_instruction_prompt_attr, render_special_instruction_prompt
    )
    workflow_prompt = get_workflow_prompt(ctx, modes_attr, render_modes)
    if conversation_history is None:
        conversation_history = ConversationHistory()
    conversation_context, new_user_message = extract_conversation_context(user_message)
    new_system_prompt = "\n".join(
        [
            make_prompt_section("Persona", persona),
            make_prompt_section("System Prompt", base_system_prompt),
            make_prompt_section("Special Instruction", special_instruction_prompt),
            make_prompt_section("Special Workflows", workflow_prompt),
            make_prompt_section(
                "Past Conversation",
                "\n".join(
                    [
                        make_prompt_section(
                            "Summary",
                            conversation_history.past_conversation_summary,
                            as_code=True,
                        ),
                        make_prompt_section(
                            "Last Transcript",
                            conversation_history.past_conversation_transcript,
                            as_code=True,
                        ),
                    ]
                ),
            ),
            make_prompt_section(
                "Notes",
                "\n".join(
                    [
                        make_prompt_section(
                            "Long Term",
                            conversation_history.long_term_note,
                            as_code=True,
                        ),
                        make_prompt_section(
                            "Contextual",
                            conversation_history.contextual_note,
                            as_code=True,
                        ),
                    ]
                ),
            ),
            make_prompt_section("Conversation Context", conversation_context),
        ]
    )
    return new_system_prompt, new_user_message


def extract_conversation_context(user_message: str) -> tuple[str, str]:
    modified_user_message = user_message
    # Match “@” + any non-space/comma sequence that contains at least one “/”
    pattern = r"(?<!\w)@(?=[^,\s]*\/)([^,\s]+)"
    potential_resource_path = re.findall(pattern, user_message)
    apendixes = []
    for i, ref in enumerate(potential_resource_path):
        resource_path = os.path.abspath(os.path.expanduser(ref))
        content = ""
        ref_type = ""
        if os.path.isfile(resource_path):
            content = read_file_with_line_numbers(resource_path)
            ref_type = "file"
        elif os.path.isdir(resource_path):
            content = read_dir(resource_path)
            ref_type = "directory"
        if content != "":
            # Replace the @-reference in the user message with the placeholder
            placeholder = f"[Reference {i+1}: {os.path.basename(ref)}]"
            modified_user_message = modified_user_message.replace(
                f"@{ref}", placeholder, 1
            )
            apendixes.append(
                make_prompt_section(
                    f"{placeholder} ({ref_type} path: `{resource_path}`)",
                    content,
                    as_code=True,
                )
            )
    conversation_context = "\n".join(
        [
            make_prompt_section("Current OS", platform.system()),
            make_prompt_section("OS Version", platform.version()),
            make_prompt_section("Python Version", platform.python_version()),
        ]
    )
    iso_date = datetime.now(timezone.utc).astimezone().isoformat()
    current_directory = os.getcwd()
    modified_user_message = "\n".join(
        [
            make_prompt_section("User Message", modified_user_message),
            make_prompt_section(
                "Context",
                "\n".join(
                    [
                        make_prompt_section(
                            "Current working directory", current_directory
                        ),
                        make_prompt_section("Current time", iso_date),
                        make_prompt_section("Apendixes", "\n".join(apendixes)),
                    ]
                ),
            ),
        ]
    )
    return conversation_context, modified_user_message


def get_user_message(
    ctx: AnyContext,
    message_attr: StrAttr | None,
    render_user_message: bool,
) -> str:
    """Gets the user message, rendering and providing a default."""
    return get_str_attr(
        ctx, message_attr, "How are you?", auto_render=render_user_message
    )


def get_summarization_system_prompt(
    ctx: AnyContext,
    summarization_prompt_attr: StrAttr | None,
    render_summarization_prompt: bool,
) -> str:
    """Gets the summarization prompt, rendering if configured and handling defaults."""
    summarization_prompt = get_attr(
        ctx,
        summarization_prompt_attr,
        None,
        auto_render=render_summarization_prompt,
    )
    if summarization_prompt is not None:
        return summarization_prompt
    return llm_config.default_summarization_prompt



import asyncio
import time
from collections import deque
from typing import Callable

import tiktoken

from zrb.config.config import CFG


def _estimate_token(text: str) -> int:
    """
    Estimates the number of tokens in a given text.
    Tries to use the 'gpt-4o' model's tokenizer for an accurate count.
    If the tokenizer is unavailable (e.g., due to network issues),
    it falls back to a heuristic of 4 characters per token.
    """
    try:
        # Primary method: Use tiktoken for an accurate count
        enc = tiktoken.encoding_for_model("gpt-4o")
        return len(enc.encode(text))
    except Exception:
        # Fallback method: Heuristic (4 characters per token)
        return len(text) // 4


class LLMRateLimiter:
    """
    Helper class to enforce LLM API rate limits and throttling.
    Tracks requests and tokens in a rolling 60-second window.
    """

    def __init__(
        self,
        max_requests_per_minute: int | None = None,
        max_tokens_per_minute: int | None = None,
        max_tokens_per_request: int | None = None,
        throttle_sleep: float | None = None,
        token_counter_fn: Callable[[str], int] | None = None,
    ):
        self._max_requests_per_minute = max_requests_per_minute
        self._max_tokens_per_minute = max_tokens_per_minute
        self._max_tokens_per_request = max_tokens_per_request
        self._throttle_sleep = throttle_sleep
        self._token_counter_fn = token_counter_fn
        self.request_times = deque()
        self.token_times = deque()

    @property
    def max_requests_per_minute(self) -> int:
        if self._max_requests_per_minute is not None:
            return self._max_requests_per_minute
        return CFG.LLM_MAX_REQUESTS_PER_MINUTE

    @property
    def max_tokens_per_minute(self) -> int:
        if self._max_tokens_per_minute is not None:
            return self._max_tokens_per_minute
        return CFG.LLM_MAX_TOKENS_PER_MINUTE

    @property
    def max_tokens_per_request(self) -> int:
        if self._max_tokens_per_request is not None:
            return self._max_tokens_per_request
        return CFG.LLM_MAX_TOKENS_PER_REQUEST

    @property
    def throttle_sleep(self) -> float:
        if self._throttle_sleep is not None:
            return self._throttle_sleep
        return CFG.LLM_THROTTLE_SLEEP

    @property
    def count_token(self) -> Callable[[str], int]:
        if self._token_counter_fn is not None:
            return self._token_counter_fn
        return _estimate_token

    def set_max_requests_per_minute(self, value: int):
        self._max_requests_per_minute = value

    def set_max_tokens_per_minute(self, value: int):
        self._max_tokens_per_minute = value

    def set_max_tokens_per_request(self, value: int):
        self._max_tokens_per_request = value

    def set_throttle_sleep(self, value: float):
        self._throttle_sleep = value

    def set_token_counter_fn(self, fn: Callable[[str], int]):
        self._token_counter_fn = fn

    def clip_prompt(self, prompt: str, limit: int) -> str:
        token_count = self.count_token(prompt)
        if token_count <= limit:
            return prompt
        while token_count > limit:
            prompt_parts = prompt.split(" ")
            last_part_index = len(prompt_parts) - 2
            clipped_prompt = " ".join(prompt_parts[:last_part_index])
            clipped_prompt += "(Content clipped...)"
            token_count = self.count_token(clipped_prompt)
            if token_count < limit:
                return clipped_prompt
        return prompt[:limit]

    async def throttle(self, prompt: str):
        now = time.time()
        tokens = self.count_token(prompt)
        # Clean up old entries
        while self.request_times and now - self.request_times[0] > 60:
            self.request_times.popleft()
        while self.token_times and now - self.token_times[0][0] > 60:
            self.token_times.popleft()
        # Check per-request token limit
        if tokens > self.max_tokens_per_request:
            raise ValueError(
                f"Request exceeds max_tokens_per_request ({self.max_tokens_per_request})."
            )
        # Wait if over per-minute request or token limit
        while (
            len(self.request_times) >= self.max_requests_per_minute
            or sum(t for _, t in self.token_times) + tokens > self.max_tokens_per_minute
        ):
            await asyncio.sleep(self.throttle_sleep)
            now = time.time()
            while self.request_times and now - self.request_times[0] > 60:
                self.request_times.popleft()
            while self.token_times and now - self.token_times[0][0] > 60:
                self.token_times.popleft()
        # Record this request
        self.request_times.append(now)
        self.token_times.append((now, tokens))


llm_rate_limitter = LLMRateLimiter()



from logging import Logger

from my_app_name.common.base_service import BaseService
from my_app_name.module.my_module.service.my_entity.repository.my_entity_repository import (
    MyEntityRepository,
)
from my_app_name.schema.my_entity import (
    MultipleMyEntityResponse,
    MyEntityCreateWithAudit,
    MyEntityResponse,
    MyEntityUpdateWithAudit,
)


class MyEntityService(BaseService):

    def __init__(self, logger: Logger, my_entity_repository: MyEntityRepository):
        super().__init__(logger)
        self.my_entity_repository = my_entity_repository

    @BaseService.route(
        "/api/v1/my-entities/{my_entity_id}",
        methods=["get"],
        response_model=MyEntityResponse,
    )
    async def get_my_entity_by_id(self, my_entity_id: str) -> MyEntityResponse:
        return await self.my_entity_repository.get_by_id(my_entity_id)

    @BaseService.route(
        "/api/v1/my-entities",
        methods=["get"],
        response_model=MultipleMyEntityResponse,
    )
    async def get_my_entities(
        self,
        page: int = 1,
        page_size: int = 10,
        sort: str | None = None,
        filter: str | None = None,
    ) -> MultipleMyEntityResponse:
        my_entities = await self.my_entity_repository.get(page, page_size, filter, sort)
        count = await self.my_entity_repository.count(filter)
        return MultipleMyEntityResponse(data=my_entities, count=count)

    @BaseService.route(
        "/api/v1/my-entities/bulk",
        methods=["post"],
        response_model=list[MyEntityResponse],
    )
    async def create_my_entity_bulk(
        self, data: list[MyEntityCreateWithAudit]
    ) -> list[MyEntityResponse]:
        my_entities = await self.my_entity_repository.create_bulk(data)
        return await self.my_entity_repository.get_by_ids(
            [my_entity.id for my_entity in my_entities]
        )

    @BaseService.route(
        "/api/v1/my-entities",
        methods=["post"],
        response_model=MyEntityResponse,
    )
    async def create_my_entity(self, data: MyEntityCreateWithAudit) -> MyEntityResponse:
        my_entity = await self.my_entity_repository.create(data)
        return await self.my_entity_repository.get_by_id(my_entity.id)

    @BaseService.route(
        "/api/v1/my-entities/bulk",
        methods=["put"],
        response_model=list[MyEntityResponse],
    )
    async def update_my_entity_bulk(
        self, my_entity_ids: list[str], data: MyEntityUpdateWithAudit
    ) -> list[MyEntityResponse]:
        await self.my_entity_repository.update_bulk(my_entity_ids, data)
        return await self.my_entity_repository.get_by_ids(my_entity_ids)

    @BaseService.route(
        "/api/v1/my-entities/{my_entity_id}",
        methods=["put"],
        response_model=MyEntityResponse,
    )
    async def update_my_entity(
        self, my_entity_id: str, data: MyEntityUpdateWithAudit
    ) -> MyEntityResponse:
        await self.my_entity_repository.update(my_entity_id, data)
        return await self.my_entity_repository.get_by_id(my_entity_id)

    @BaseService.route(
        "/api/v1/my-entities/bulk",
        methods=["delete"],
        response_model=list[MyEntityResponse],
    )
    async def delete_my_entity_bulk(
        self, my_entity_ids: list[str], deleted_by: str
    ) -> list[MyEntityResponse]:
        my_entities = await self.my_entity_repository.get_by_ids(my_entity_ids)
        await self.my_entity_repository.delete_bulk(my_entity_ids)
        return my_entities

    @BaseService.route(
        "/api/v1/my-entities/{my_entity_id}",
        methods=["delete"],
        response_model=MyEntityResponse,
    )
    async def delete_my_entity(
        self, my_entity_id: str, deleted_by: str
    ) -> MyEntityResponse:
        my_entity = await self.my_entity_repository.get_by_id(my_entity_id)
        await self.my_entity_repository.delete(my_entity_id)
        return my_entity



from zrb.builtin.group import shell_autocomplete_group
from zrb.context.context import Context
from zrb.runner.cli import cli
from zrb.task.make_task import make_task
from zrb.util.cli.subcommand import get_group_subcommands


@make_task(
    name="get-shell-subcommands",
    description="Get subcommand of any Zrb command",
    group=shell_autocomplete_group,
    alias="subcmd",
)
def get_shell_subcommands(ctx: Context):
    subcommands = get_group_subcommands(cli)
    for subcommand in subcommands:
        if subcommand.paths == ctx.args:
            return " ".join(subcommand.nexts)
    return ""



# No specific imports needed from typing for these changes
from zrb.task.any_task import AnyTask


def handle_rshift(
    left_task: AnyTask, right_operand: AnyTask | list[AnyTask]
) -> AnyTask | list[AnyTask]:
    """
    Implements the >> operator logic: left_task becomes an upstream for right_operand.
    Modifies the right_operand(s) by calling append_upstream.
    Returns the right_operand.
    """
    try:
        if isinstance(right_operand, list):
            for task in right_operand:
                # Assuming append_upstream exists and handles duplicates
                task.append_upstream(left_task)
        else:
            # Assuming right_operand is a single AnyTask
            right_operand.append_upstream(left_task)
        return right_operand
    except Exception as e:
        # Catch potential errors during append_upstream or type issues
        raise ValueError(f"Invalid operation {left_task} >> {right_operand}: {e}")


def handle_lshift(
    left_task: AnyTask, right_operand: AnyTask | list[AnyTask]
) -> AnyTask:
    """
    Implements the << operator logic: right_operand becomes an upstream for left_task.
    Modifies the left_task by calling append_upstream.
    Returns the left_task.
    """
    try:
        # Assuming append_upstream exists and handles single or list input
        left_task.append_upstream(right_operand)
        return left_task
    except Exception as e:
        # Catch potential errors during append_upstream or type issues
        raise ValueError(f"Invalid operation {left_task} << {right_operand}: {e}")



import datetime

import ulid
from pydantic import BaseModel
from sqlmodel import Field, SQLModel


class MyEntityBase(SQLModel):
    my_column: str


class MyEntityCreate(MyEntityBase):
    def with_audit(self, created_by: str) -> "MyEntityCreateWithAudit":
        return MyEntityCreateWithAudit(**self.model_dump(), created_by=created_by)


class MyEntityCreateWithAudit(MyEntityCreate):
    created_by: str


class MyEntityUpdate(SQLModel):
    my_column: str | None = None

    def with_audit(self, updated_by: str) -> "MyEntityUpdateWithAudit":
        return MyEntityUpdateWithAudit(
            **self.model_dump(exclude_none=True), updated_by=updated_by
        )


class MyEntityUpdateWithAudit(MyEntityUpdate):
    updated_by: str


class MyEntityResponse(MyEntityBase):
    id: str


class MultipleMyEntityResponse(BaseModel):
    data: list[MyEntityResponse]
    count: int


class MyEntity(SQLModel, table=True):
    __tablename__ = "my_entities"
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    created_at: datetime.datetime = Field(index=True)
    created_by: str = Field(index=True)
    updated_at: datetime.datetime | None = Field(index=True)
    updated_by: str | None = Field(index=True)
    my_column: str = Field(index=True)



import json
import traceback
from typing import TYPE_CHECKING

from zrb.attr.type import BoolAttr, IntAttr
from zrb.config.llm_config import llm_config
from zrb.config.llm_rate_limitter import LLMRateLimiter, llm_rate_limitter
from zrb.context.any_context import AnyContext
from zrb.task.llm.agent import run_agent_iteration
from zrb.task.llm.conversation_history import (
    count_part_in_history_list,
    replace_system_prompt_in_history,
)
from zrb.task.llm.conversation_history_model import ConversationHistory
from zrb.task.llm.typing import ListOfDict
from zrb.util.attr import get_bool_attr, get_int_attr
from zrb.util.cli.style import stylize_faint
from zrb.util.llm.prompt import make_prompt_section

if TYPE_CHECKING:
    from pydantic_ai.models import Model
    from pydantic_ai.settings import ModelSettings


def _count_token_in_history(history_list: ListOfDict) -> int:
    """Counts the total number of tokens in a conversation history list."""
    text_to_count = json.dumps(history_list)
    return llm_rate_limitter.count_token(text_to_count)


def get_history_summarization_token_threshold(
    ctx: AnyContext,
    history_summarization_token_threshold_attr: IntAttr | None,
    render_history_summarization_token_threshold: bool,
) -> int:
    """Gets the history summarization token threshold, handling defaults and errors."""
    try:
        return get_int_attr(
            ctx,
            history_summarization_token_threshold_attr,
            llm_config.default_history_summarization_token_threshold,
            auto_render=render_history_summarization_token_threshold,
        )
    except ValueError as e:
        ctx.log_warning(
            f"Could not convert history_summarization_token_threshold to int: {e}. "
            "Defaulting to -1 (no threshold)."
        )
        return -1


def should_summarize_history(
    ctx: AnyContext,
    history_list: ListOfDict,
    should_summarize_history_attr: BoolAttr | None,
    render_summarize_history: bool,
    history_summarization_token_threshold_attr: IntAttr | None,
    render_history_summarization_token_threshold: bool,
) -> bool:
    """Determines if history summarization should occur based on token length and config."""
    history_part_count = count_part_in_history_list(history_list)
    if history_part_count == 0:
        return False
    summarization_token_threshold = get_history_summarization_token_threshold(
        ctx,
        history_summarization_token_threshold_attr,
        render_history_summarization_token_threshold,
    )
    history_token_count = _count_token_in_history(history_list)
    if (
        summarization_token_threshold == -1
        or summarization_token_threshold > history_token_count
    ):
        return False
    return get_bool_attr(
        ctx,
        should_summarize_history_attr,
        llm_config.default_summarize_history,
        auto_render=render_summarize_history,
    )


async def summarize_history(
    ctx: AnyContext,
    model: "Model | str | None",
    settings: "ModelSettings | None",
    system_prompt: str,
    conversation_history: ConversationHistory,
    rate_limitter: LLMRateLimiter | None = None,
    retries: int = 3,
) -> str:
    """Runs an LLM call to update the conversation summary."""
    from pydantic_ai import Agent

    ctx.log_info("Attempting to summarize conversation history...")
    # Construct the user prompt for the summarization agent
    user_prompt = "\n".join(
        [
            make_prompt_section(
                "Past Conversation",
                "\n".join(
                    [
                        make_prompt_section(
                            "Summary",
                            conversation_history.past_conversation_summary,
                            as_code=True,
                        ),
                        make_prompt_section(
                            "Last Transcript",
                            conversation_history.past_conversation_transcript,
                            as_code=True,
                        ),
                    ]
                ),
            ),
            make_prompt_section(
                "Recent Conversation (JSON)",
                json.dumps(conversation_history.history),
                as_code=True,
            ),
            make_prompt_section(
                "Notes",
                "\n".join(
                    [
                        make_prompt_section(
                            "Long Term",
                            conversation_history.long_term_note,
                            as_code=True,
                        ),
                        make_prompt_section(
                            "Contextual",
                            conversation_history.contextual_note,
                            as_code=True,
                        ),
                    ]
                ),
            ),
        ]
    )
    summarization_agent = Agent(
        model=model,
        system_prompt=system_prompt,
        model_settings=settings,
        retries=retries,
        tools=[
            conversation_history.write_past_conversation_summary,
            conversation_history.write_past_conversation_transcript,
            conversation_history.read_long_term_note,
            conversation_history.write_long_term_note,
            conversation_history.read_contextual_note,
            conversation_history.write_contextual_note,
        ],
    )
    try:
        ctx.print(stylize_faint("📝 Summarize Conversation >>>"), plain=True)
        summary_run = await run_agent_iteration(
            ctx=ctx,
            agent=summarization_agent,
            user_prompt=user_prompt,
            history_list=[],
            rate_limitter=rate_limitter,
        )
        if summary_run and summary_run.result and summary_run.result.output:
            usage = summary_run.result.usage()
            ctx.print(stylize_faint(f"📝 Summarization Token: {usage}"), plain=True)
            ctx.print(plain=True)
            ctx.log_info("History summarized and updated.")
        else:
            ctx.log_warning("History summarization failed or returned no data.")
    except BaseException as e:
        ctx.log_warning(f"Error during history summarization: {e}")
        traceback.print_exc()
    # Return the original summary if summarization fails
    return conversation_history


async def maybe_summarize_history(
    ctx: AnyContext,
    conversation_history: ConversationHistory,
    should_summarize_history_attr: BoolAttr | None,
    render_summarize_history: bool,
    history_summarization_token_threshold_attr: IntAttr | None,
    render_history_summarization_token_threshold: bool,
    model: "str | Model | None",
    model_settings: "ModelSettings | None",
    summarization_prompt: str,
    rate_limitter: LLMRateLimiter | None = None,
) -> ConversationHistory:
    """Summarizes history and updates context if enabled and threshold met."""
    shorten_history = replace_system_prompt_in_history(conversation_history.history)
    if should_summarize_history(
        ctx,
        shorten_history,
        should_summarize_history_attr,
        render_summarize_history,
        history_summarization_token_threshold_attr,
        render_history_summarization_token_threshold,
    ):
        original_history = conversation_history.history
        conversation_history.history = shorten_history
        conversation_history = await summarize_history(
            ctx=ctx,
            model=model,
            settings=model_settings,
            system_prompt=summarization_prompt,
            conversation_history=conversation_history,
            rate_limitter=rate_limitter,
        )
        conversation_history.history = original_history
        if (
            conversation_history.past_conversation_summary != ""
            and conversation_history.past_conversation_transcript != ""
        ):
            conversation_history.history = []
    return conversation_history



from my_app_name.common.db_engine_factory import db_engine
from my_app_name.config import APP_REPOSITORY_TYPE
from my_app_name.module.auth.service.permission.repository.permission_db_repository import (
    PermissionDBRepository,
)
from my_app_name.module.auth.service.permission.repository.permission_repository import (
    PermissionRepository,
)

if APP_REPOSITORY_TYPE == "db":
    permission_repository: PermissionRepository = PermissionDBRepository(db_engine)
else:
    permission_repository: PermissionRepository = None
